﻿using System;
using System.Collections.Generic;
using System.Text;
using System.IO.Ports;
using System.Runtime.InteropServices;

namespace RX7Interface
{
    class Rx7SerialPort
    {
        private const int BaudRate = 976;
        private uint timerId = uint.MaxValue;

        private SerialPort serialPort;

        private byte[] lastSentPacket;
        private int loopbackPosition;

        private List<byte> receiveBuffer = new List<byte>();

        private MMTimerProc timerHandle = null;

        private delegate void MMTimerProc(UInt32 timerid, UInt32 msg, IntPtr user, UInt32 dw1, UInt32 dw2);

        [DllImport("winmm.dll")]
        static extern uint timeSetEvent(
                  UInt32 uDelay,
                  UInt32 uResolution,
                  [MarshalAs(UnmanagedType.FunctionPtr)] MMTimerProc lpTimeProc,
                  UInt32 dwUser,
                  Int32 fuEvent
                );

        [DllImport("winmm.dll")]
        static extern uint timeKillEvent(uint uTimerID);

        public Rx7SerialPort(string port)
        {
            serialPort = new SerialPort(port, BaudRate, Parity.None, 8, StopBits.One);
        }

        /// <summary>
        /// Gets the number of available bytes.
        /// </summary>
        public int BytesAvailable
        {
            get
            {
                ProcessReceiveBuffer();
                return receiveBuffer.Count;
            }
        }

        /// <summary>
        /// Transmits data to the ECU.
        /// </summary>
        /// <param name="data"></param>
        public void Write(byte[] data)
        {
            // Clear any old received data, as the loopback will screw the data up.
            serialPort.DiscardInBuffer();

            lastSentPacket = data;
            loopbackPosition = 0;
            serialPort.Write(data, 0, data.Length);
        }

        /// <summary>
        /// Gets all available data that was received, with and loopback
        /// data excluded.
        /// </summary>
        /// <returns></returns>
        public byte[] Read()
        {
            ProcessReceiveBuffer();

            byte[] data = receiveBuffer.ToArray();

            receiveBuffer.Clear();

            return data;
        }

        public void Open()
        {
            serialPort.Open();
            StartMultimediaTimer();
        }

        public void Close()
        {
            StopMultimediaTimer();
            serialPort.Close();
        }

        public bool IsOpen
        {
            get
            {
                return serialPort.IsOpen;
            }
        }

        private void ProcessReceiveBuffer()
        {
            if (serialPort.BytesToRead > 0)
            {
                if (lastSentPacket != null)
                {
                    // Remove the loopback bytes;
                    while (loopbackPosition < lastSentPacket.Length && serialPort.BytesToRead > 0)
                    {
                        int byteRead = serialPort.ReadByte();
                        if (lastSentPacket[loopbackPosition] != byteRead)
                        {
                            // Loopback data does not match.
                            Console.WriteLine("Loopback packet doesn't match.");
                            lastSentPacket = null;
                            break;
                        }

                        loopbackPosition++;
                    }
                }

                while (serialPort.BytesToRead > 0)
                {
                    receiveBuffer.Add((byte)serialPort.ReadByte());
                }
            }
        }

        // Callback from the deterministic multimedia timer.
        private void MultimediaTimerCallback(UInt32 timerid, UInt32 msg, IntPtr user, UInt32 dw1, UInt32 dw2)
        {
            serialPort.RtsEnable = !serialPort.RtsEnable;
        }

        private void StartMultimediaTimer()
        {
            if (timerId == uint.MaxValue)
            {
                timerHandle = new MMTimerProc(MultimediaTimerCallback);
                timerId = timeSetEvent(25, 0, timerHandle, 0, 1);
            }
        }

        private void StopMultimediaTimer()
        {
            if (timerId != uint.MaxValue)
            {
                timeKillEvent(timerId);
                timerId = uint.MaxValue;
            }
        }
    }
}
