﻿using System;
using System.Text;
using System.Runtime.InteropServices;
using System.Collections.Generic;
using System.Threading;
using System.IO;
using System.Drawing;
using System.Windows.Forms;
using RX7Interface.Gauges;

namespace RX7Interface
{
    public partial class MainForm : Form
    {
        private const string fileName = "log";

        private TextWriter logFileWriter;

        private bool readingParameters = true;
        private bool loggingActive = false;

        private IDataStream dataStream;

        private Parameter[] parameters = new Parameter[] 
        {
            new Parameter("Intake Air Pressure", "mmHg", new DataValue(0x0021, ParameterLength.TwoBytes, 500d/256, -1000), null),
            new Parameter("Throttle Angle (Narrow)", "V", new DataValue(0x0023, ParameterLength.OneByte, 5d/256, 0), new GaugeInfo(0, 5)),
            new Parameter("Throttle Angle (Wide)", "V", new DataValue(0x0024, ParameterLength.OneByte, 5d/256, 0), new GaugeInfo(0, 5)),
            new Parameter("Oxygen Sensor Voltage", "V",new DataValue(0x0025, ParameterLength.OneByte, 5d/256, 0), new GaugeInfo(0, 5)),
            new Parameter("MOP Position", "V", new DataValue(0x0026, ParameterLength.OneByte, 5d/256, 0), null),
            new Parameter("Battery Voltage", "V", new DataValue(0x0027, ParameterLength.OneByte, 20d/256, 0), new GaugeInfo(8, 20, 11, 15)),
            new Parameter("Water Temperature", "°C", new DataValue(0x0028, ParameterLength.OneByte, 160d/256, -40), new GaugeInfo(0, 140, 80, 100)),
            new Parameter("Fuel Temperature", "°C", new DataValue(0x0029, ParameterLength.OneByte, 160d/256, 0), null),
            new Parameter("Intake Air Temperature", "°C",new DataValue(0x002a, ParameterLength.OneByte, 160d/256, -40), null),
            new Parameter("Engine Speed", "rpm", new DataValue(0x002c, ParameterLength.TwoBytes, 500d/256, 0), new GaugeInfo(0, 9000)),
            new Parameter("Vehicle Speed", "km/h", new DataValue(0x002e, ParameterLength.OneByte, 356d/256, 0), new GaugeInfo(0, 320)),
            new Parameter("Injector On Time", "ms", new DataValue(0x0800, ParameterLength.TwoBytes, 1d/256, 0), null),
            new Parameter("Injector Period", "ms", new DataValue(0x0802, ParameterLength.TwoBytes, 1d/256, 0), null),
            new Parameter("Ignition Advance (Leading)", "degrees", new DataValue(0x0804, ParameterLength.OneByte, 90d/256, -25), null),
            new Parameter("Ignition Advance (Trailling)", "degrees", new DataValue(0x0805, ParameterLength.OneByte, 90d/256, -25), null),
            new Parameter("Idle Speed Control", "%", new DataValue(0x0806, ParameterLength.TwoBytes, 800d/256, 0), null),
            new Parameter("Unknown 1", "?", new DataValue(0x0808, ParameterLength.OneByte, 1, 0), null),
            new Parameter("Turbo Pre-Control", "%", new DataValue(0x0809, ParameterLength.OneByte, 100d/256, 0), null),
            new Parameter("Wastegate Control", "%", new DataValue(0x080a, ParameterLength.OneByte, 100d/256, 0), null),

            
            new Parameter("Unknown 2", "?", new DataValue(0x080b, ParameterLength.OneByte, 1, 0), null),
            new Parameter("Unknown 3", "?", new DataValue(0x080c, ParameterLength.OneByte, 1, 0), null),
            new Parameter("Unknown 4", "?", new DataValue(0x080d, ParameterLength.OneByte, 1, 0), null),
            new Parameter("Unknown 5", "?", new DataValue(0x080e, ParameterLength.OneByte, 1, 0), null),
            new Parameter("Unknown 6", "?", new DataValue(0x080f, ParameterLength.OneByte, 1, 0), null),
            new Parameter("Unknown 7", "?", new DataValue(0x0810, ParameterLength.OneByte, 1, 0), null),
            new Parameter("Unknown 8", "?", new DataValue(0x0811, ParameterLength.OneByte, 1, 0), null),
            new Parameter("Unknown 9", "?", new DataValue(0x0812, ParameterLength.OneByte, 1, 0), null),
            new Parameter("Unknown 10", "?", new DataValue(0x0813, ParameterLength.OneByte, 1, 0), null)
        };

        /* Remaining Paramters (addresses unknown):
        MOP 	Metering Oil Pump Position 	V
        TPC 	Turbo Pre-Control Valve 	%
        WGC 	Wastegate Control Valve 	%
        CRFSV 	Charge Relief Valve 	ON . OFF
        CCNTSV 	Charge Control Valve 	ON . OFF
        TCNTSV 	Turbo Control Valve 	ON . OFF
        RAD-FAN1 	Radiator Fan Relay 1 	ON . OFF
        RAD-FAN2 	Radiator Fan Relay 2 	ON . OFF
        RAD-FAN3 	Radiator Fan Relay 3 	ON . OFF
        AC-SW   	AC Switch  	ON . OFF 
        AC-RLY 	AC Relay 	ON . OFF
        PRCSV 	Pressure Regulator Control Valve 	ON . OFF
        DTCNTSV 	Double Throttle Control Valve 	ON . OFF
        FP-RLY 	Fuel Pump Relay 	ON . OFF
        Diagnostic Code 	OK.NG CODE
         */

        public MainForm()
        {
            InitializeComponent();
        }

        private string GetLogFileName()
        {
            return fileName + " " +  DateTime.Now.ToString("dd-mm-yy HH-mm") + ".txt";
        }

        private GaugeForm gaugeForm;
        
        private void buttonOpenClose_Click(object sender, EventArgs e)
        {
            if (dataStream == null || !dataStream.IsOpen && comboPorts.SelectedItem != null)
            {
                //dataStream = new DummyDataStream();
                dataStream = new Rx7DataStream(comboPorts.SelectedItem as string);
                dataStream.Open();

                buttonOpenClose.Text = "Close";

                gaugeForm = new GaugeForm();
                gaugeForm.Show();
                gaugeForm.SetParameterList(parameters);

                backgroundWorker.RunWorkerAsync();
            }
            else if (dataStream != null && dataStream.IsOpen)
            {
                logFileWriter.Dispose();
                dataStream.Close();
                buttonOpenClose.Text = "Open";
            }
        }

        private void Form1_Load(object sender, EventArgs e)
        {
            foreach(string port in System.IO.Ports.SerialPort.GetPortNames())
            {
                comboPorts.Items.Add(port);
            }

            comboPorts.SelectedIndex = comboPorts.Items.Count - 1;

            bool alternating = true;

            foreach (Parameter p in parameters)
            {
                listViewParameters.Items.Add(p.Name, p.Name, 0);

                ListViewItem item = listViewParameters.Items[p.Name];

                item.BackColor = alternating ? Color.GhostWhite : Color.White;
                alternating = !alternating;
                item.Tag = p;
                item.Checked = p.Enabled;
                item.SubItems.AddRange(new string[]
                {
                    "-",
                    p.Units,
                    "0x" + p.DataValue.Addresses[0].ToString("X4"),
                    p.DataValue.Addresses.Length.ToString(),
                    "-"                    
                });
            }

            timerUpdate.Start();

            //buttonOpenClose_Click(null, null);
        }

        private void MainForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (gaugeForm != null)
            {
                gaugeForm.Close();
            }

            if (logFileWriter != null)
            {
                logFileWriter.Dispose();
            }

            backgroundWorker.CancelAsync();
        }

        private void backgroundWorker_DoWork(object sender, System.ComponentModel.DoWorkEventArgs e)
        {
            while (!e.Cancel)
            {
                try
                {
                    if (readingParameters)
                    {
                        foreach (Parameter p in parameters)
                        {
                            if (p.Enabled)
                            {
                                foreach (uint address in p.DataValue.Addresses)
                                {
                                    byte received = dataStream.ReadByte(address);
                                    p.DataValue.SetRawValue(address, received);
                                    p.DataValue.UpdateTime = DateTime.Now;
                                    backgroundWorker.ReportProgress(0);
                                }

                                if (loggingActive)
                                {
                                    LogParameter(p);
                                }
                            }
                        }
                    }
                    else
                    {
                        Thread.Sleep(50);
                    }
                }
                catch
                {
                }
            }
        }

        private void LogParameter(Parameter p)
        {
            if (logFileWriter == null)
            {
                logFileWriter = new StreamWriter(GetLogFileName(), true);
            }

            logFileWriter.WriteLine(string.Format("{0}\t{1}\t{2}\t{3}", DateTime.Now.ToString("HH:mm:ss.ff"), p.Name, p.DataValue.GetValue(), p.Units));
        }

        private void backgroundWorker_ProgressChanged(object sender, System.ComponentModel.ProgressChangedEventArgs e)
        {
            listViewParameters.SuspendLayout();

            foreach (Parameter p in parameters)
            {
                listViewParameters.Items[p.Name].SubItems[1].Text = p.DataValue.GetValue().ToString("f2");
            }

            listViewParameters.ResumeLayout();

            if (gaugeForm != null && !gaugeForm.IsDisposed && gaugeForm.Visible)
            {
                gaugeForm.Update();
            }
        }

        private void backgroundWorker_RunWorkerCompleted(object sender, System.ComponentModel.RunWorkerCompletedEventArgs e)
        {
            dataStream.Close();
        }

        private void checkAutoRead_CheckedChanged(object sender, EventArgs e)
        {
            readingParameters = checkAutoRead.Checked;
        }

        private void listViewParameters_ItemChecked(object sender, ItemCheckedEventArgs e)
        {
            Parameter parameter = e.Item.Tag as Parameter;

            if (parameter != null)
            {
                parameter.Enabled = e.Item.Checked;
            }
        }

        private void timerUpdate_Tick(object sender, EventArgs e)
        {
            foreach (ListViewItem item in listViewParameters.Items)
            {
                Parameter parameter = item.Tag as Parameter;
                if (parameter.DataValue.UpdateTime != DateTime.MinValue)
                {
                    TimeSpan age = DateTime.Now - parameter.DataValue.UpdateTime;
                    item.SubItems[5].Text = age.TotalSeconds.ToString("f0");
                }
            }
        }

        private void checkLog_CheckedChanged(object sender, EventArgs e)
        {
            loggingActive = checkLog.Checked;
        }

        private void buttonCheckAll_Click(object sender, EventArgs e)
        {
            foreach (ListViewItem item in listViewParameters.Items)
            {
                item.Checked = true;
            }
        }

        private void buttonCheckNone_Click(object sender, EventArgs e)
        {
            foreach (ListViewItem item in listViewParameters.Items)
            {
                item.Checked = false;
            }
        }
    }
}
